import numpy as np
import dask.array as dsa
from ..duck_array_ops import (
    digitize,
    bincount,
    reshape,
    ravel_multi_index,
    broadcast_arrays,
)
from .fixtures import empty_dask_array
import pytest


@pytest.mark.parametrize(
    "function, args",
    [
        (digitize, [np.random.rand(5, 12), np.linspace(0, 1, 7)]),
        (bincount, [np.arange(10)]),
    ],
)
def test_eager(function, args):
    a = function(*args)
    assert isinstance(a, np.ndarray)


@pytest.mark.parametrize(
    "function, args, kwargs",
    [
        (digitize, [empty_dask_array((5, 12)), np.linspace(0, 1, 7)], {}),
        (bincount, [empty_dask_array((10,))], {"minlength": 5}),
        (reshape, [empty_dask_array((10, 5)), (5, 10)], {}),
        (ravel_multi_index, (empty_dask_array((10,)), empty_dask_array((10,))), {}),
    ],
)
def test_lazy(function, args, kwargs):
    # make sure nothing computes
    a = function(*args, **kwargs)
    assert isinstance(a, dsa.core.Array)


@pytest.mark.parametrize("chunks", [(5, 12), (1, 12), (5, 1)])
def test_digitize_dask_correct(chunks):
    a = np.random.rand(5, 12)
    da = dsa.from_array(a, chunks=chunks)
    bins = np.linspace(0, 1, 7)
    d = digitize(a, bins)
    dd = digitize(da, bins)
    np.testing.assert_array_equal(d, dd.compute())


def test_ravel_multi_index_correct():
    arr = np.array([[3, 6, 6], [4, 5, 1]])
    expected = np.ravel_multi_index(arr, (7, 6))
    actual = ravel_multi_index(arr, (7, 6))
    np.testing.assert_array_equal(expected, actual)

    expected = np.ravel_multi_index(arr, (7, 6), order="F")
    actual = ravel_multi_index(arr, (7, 6), order="F")
    np.testing.assert_array_equal(expected, actual)


def test_broadcast_arrays_numpy():
    a1 = np.empty((1, 5, 25))
    a2 = np.empty((4, 1, 1))

    a1b, a2b = broadcast_arrays(a1, a2)
    assert a1b.shape == (4, 5, 25)
    assert a2b.shape == (4, 5, 25)


@pytest.mark.parametrize("d1_chunks", [(5 * (1,), (25,)), ((2, 3), (25,))])
def test_broadcast_arrays_dask(d1_chunks):
    d1 = dsa.empty((5, 25), chunks=d1_chunks)
    d2 = dsa.empty((1, 25), chunks=(1, 25))

    d1b, d2b = broadcast_arrays(d1, d2)
    assert d1b.shape == (5, 25)
    assert d2b.shape == (5, 25)
    assert d1b.chunks == d1_chunks
    assert d2b.chunks == d1_chunks



import dask
import dask.array as dsa


def empty_dask_array(shape, dtype=float, chunks=None):
    # a dask array that errors if you try to comput it
    def raise_if_computed():
        raise ValueError("Triggered forbidden computation")

    a = dsa.from_delayed(dask.delayed(raise_if_computed)(), shape, dtype)
    if chunks is not None:
        a = a.rechunk(chunks)

    return a



import xarray as xr
import numpy as np
import pytest
import pandas as pd
from itertools import combinations

from ..xarray import histogram


# example dimensions
DIMS = {"time": 5, "depth": 10, "lat": 45, "lon": 90}
COORDS = {
    "time": ("time", pd.date_range(start="2000-01-01", periods=DIMS["time"])),
    "depth": ("depth", np.arange(DIMS["depth"]) * 100.0 + 50),
    "lat": ("lat", np.arange(DIMS["lat"]) * 180 / DIMS["lat"] - 90 + 90 / DIMS["lat"]),
    "lon": ("lon", np.arange(DIMS["lon"]) * 360 / DIMS["lon"] + 180 / DIMS["lon"]),
}


@pytest.fixture(
    params=[
        ("lon",),
        ("lat", "lon"),
        ("depth", "lat", "lon"),
        ("time", "depth", "lat", "lon"),
    ],
    ids=["1D", "2D", "3D", "4D"],
)
def ones(request):
    dims = request.param
    shape = [DIMS[d] for d in dims]
    coords = {k: v for k, v in COORDS.items() if k in dims}
    data = np.ones(shape, dtype="f8")
    da = xr.DataArray(data, dims=dims, coords=coords, name="ones")
    return da


@pytest.mark.parametrize("ndims", [1, 2, 3, 4])
def test_histogram_ones(ones, ndims):
    dims = ones.dims
    if ones.ndim < ndims:
        pytest.skip(
            "Don't need to test when number of dimension combinations "
            "exceeds the number of array dimensions"
        )

    # everything should be in the middle bin (index 1)
    bins = np.array([0, 0.9, 1.1, 2])
    bins_c = 0.5 * (bins[1:] + bins[:-1])

    def _check_result(h, d):
        other_dims = [dim for dim in ones.dims if dim not in d]
        if len(other_dims) > 0:
            assert set(other_dims) <= set(h.dims)
        # check that all values are in the central bin
        h_sum = h.sum(other_dims)
        h_sum_expected = xr.DataArray(
            [0, ones.size, 0],
            dims=["ones_bin"],
            coords={"ones_bin": ("ones_bin", bins_c)},
            name="histogram_ones",
        )
        xr.testing.assert_identical(h_sum, h_sum_expected)

    for d in combinations(dims, ndims):
        h = histogram(ones, bins=[bins], dim=d)
        _check_result(h, d)


@pytest.mark.parametrize("ndims", [1, 2, 3, 4])
def test_histogram_ones_density(ones, ndims):
    dims = ones.dims
    if ones.ndim < ndims:
        pytest.skip(
            "Don't need to test when number of dimension combinations "
            "exceeds the number of array dimensions"
        )

    # everything should be in the middle bin (index 1)
    bins = np.array([0, 0.9, 1.1, 2])
    bin_area = 0.2

    def _check_result(h_density, d):
        other_dims = [dim for dim in ones.dims if dim not in d]
        if len(other_dims) > 0:
            assert set(other_dims) <= set(h_density.dims)

        # check that all integrals over pdfs at different locations are = 1
        h_integrals = (h_density * bin_area).sum(dim="ones_bin")
        np.testing.assert_allclose(h_integrals.values, 1.0)

    for d in combinations(dims, ndims):
        h_density = histogram(ones, bins=[bins], dim=d, density=True)
        _check_result(h_density, d)


# TODO: refactor this test to use better fixtures
# (it currently has a ton of loops)
@pytest.mark.parametrize("ndims", [1, 2, 3, 4])
def test_weights(ones, ndims):
    dims = ones.dims
    if ones.ndim < ndims:
        pytest.skip(
            "Don't need to test when number of dimension combinations "
            "exceeds the number of array dimensions"
        )

    bins = np.array([0, 0.9, 1.1, 2])
    bins_c = 0.5 * (bins[1:] + bins[:-1])

    weight_value = 0.5

    def _check_result(h, d):
        other_dims = [dim for dim in ones.dims if dim not in d]
        if len(other_dims) > 0:
            assert set(other_dims) <= set(h.dims)
        # check that all values are in the central bin
        h_sum = h.sum(other_dims)
        h_sum_expected = xr.DataArray(
            [0, weight_value * ones.size, 0],
            dims=["ones_bin"],
            coords={"ones_bin": ("ones_bin", bins_c)},
            name="histogram_ones",
        )
        xr.testing.assert_identical(h_sum, h_sum_expected)

    # get every possible combination of sub-dimensions
    for n_combinations in range(ones.ndim):
        for weight_dims in combinations(dims, n_combinations):
            i_selector = {dim: 0 for dim in weight_dims}
            weights = xr.full_like(ones.isel(**i_selector), weight_value)
            for nc in range(ndims):
                for d in combinations(dims, nc + 1):
                    h = histogram(ones, weights=weights, bins=[bins], dim=d)
                    _check_result(h, d)


# test for issue #5
def test_dims_and_coords():
    time_axis = np.arange(4)
    depth_axis = np.arange(10)
    X_axis = np.arange(30)
    Y_axis = np.arange(30)

    dat1 = np.random.randint(
        low=0,
        high=100,
        size=(len(time_axis), len(depth_axis), len(X_axis), len(Y_axis)),
    )
    array1 = xr.DataArray(
        dat1,
        coords=[time_axis, depth_axis, X_axis, Y_axis],
        dims=["time", "depth", "X", "Y"],
        name="one",
    )

    dat2 = np.random.randint(
        low=0, high=50, size=(len(time_axis), len(depth_axis), len(X_axis), len(Y_axis))
    )
    array2 = xr.DataArray(
        dat2,
        coords=[time_axis, depth_axis, X_axis, Y_axis],
        dims=["time", "depth", "X", "Y"],
        name="two",
    )

    bins1 = np.linspace(0, 100, 50)
    bins2 = np.linspace(0, 50, 25)

    result = histogram(array1, array2, dim=["X", "Y"], bins=[bins1, bins2])
    assert result.dims == ("time", "depth", "one_bin", "two_bin")
    assert result.time.identical(array1.time)
    assert result.depth.identical(array2.depth)


@pytest.mark.parametrize("number_of_inputs", [1, 2])
@pytest.mark.parametrize("keep_coords", [True, False])
def test_carry_coords(keep_coords, number_of_inputs):
    time_axis = np.arange(40)
    X_axis = np.arange(10)
    Y_axis = np.arange(10)
    data = np.random.randint(
        low=0, high=100, size=(len(time_axis), len(X_axis), len(Y_axis))
    )
    da = xr.DataArray(
        data, coords=[time_axis, X_axis, Y_axis], dims=["time", "X", "Y"], name="one"
    )
    # faking coordinates
    da["lon"] = da.X ** 2 + da.Y ** 2
    assert "lon" in da.coords
    bins = np.linspace(0, 100, 10)
    result = histogram(
        *[da] * number_of_inputs,
        bins=[bins] * number_of_inputs,
        dim=["time"],
        keep_coords=keep_coords
    )
    if keep_coords:
        assert "lon" in result.coords
    else:
        assert "lon" not in result.coords



"""
Xarray API for xhistogram.
"""

import xarray as xr
import numpy as np
from collections import OrderedDict
from .core import histogram as _histogram
import warnings


def histogram(
    *args,
    bins=None,
    dim=None,
    weights=None,
    density=False,
    block_size="auto",
    keep_coords=False,
    bin_dim_suffix="_bin",
    bin_edge_suffix="_bin_edge"
):
    """Histogram applied along specified dimensions.

    Parameters
    ----------
    args : xarray.DataArray objects
        Input data. The number of input arguments determines the dimensonality of
        the histogram. For example, two arguments prodocue a 2D histogram. All
        args must be aligned and have the same dimensions.
    bins :  int or array_like or a list of ints or arrays, optional
        If a list, there should be one entry for each item in ``args``.
        The bin specification:

          * If int, the number of bins for all arguments in ``args``.
          * If array_like, the bin edges for all arguments in ``args``.
          * If a list of ints, the number of bins  for every argument in ``args``.
          * If a list arrays, the bin edges for each argument in ``args``
            (required format for Dask inputs).
          * A combination [int, array] or [array, int], where int
            is the number of bins and array is the bin edges.

        When bin edges are specified, all but the last (righthand-most) bin include
        the left edge and exclude the right edge. The last bin includes both edges.

        A ``TypeError`` will be raised if ``args`` contains dask arrays and
        ``bins`` are not specified explicitly as a list of arrays.
    dim : tuple of strings, optional
        Dimensions over which which the histogram is computed. The default is to
        compute the histogram of the flattened array.
    weights : array_like, optional
        An array of weights, of the same shape as `a`.  Each value in
        `a` only contributes its associated weight towards the bin count
        (instead of 1). If `density` is True, the weights are
        normalized, so that the integral of the density over the range
        remains 1. NaNs in the weights input will fill the entire bin with
        NaNs. If there are NaNs in the weights input call ``.fillna(0.)``
        before running ``histogram()``.
    density : bool, optional
        If ``False``, the result will contain the number of samples in
        each bin. If ``True``, the result is the value of the
        probability *density* function at the bin, normalized such that
        the *integral* over the range is 1. Note that the sum of the
        histogram values will not be equal to 1 unless bins of unity
        width are chosen; it is not a probability *mass* function.
    block_size : int or 'auto', optional
        A parameter which governs the algorithm used to compute the histogram.
        Using a nonzero value splits the histogram calculation over the
        non-histogram axes into blocks of size ``block_size``, iterating over
        them with a loop (numpy inputs) or in parallel (dask inputs). If
        ``'auto'``, blocks will be determined either by the underlying dask
        chunks (dask inputs) or an experimental built-in heuristic (numpy inputs).
    keep_coords : bool, optional
        If ``True``, keep all coordinates. Default: ``False``

    Returns
    -------
    hist : array
        The values of the histogram.

    """

    N_args = len(args)

    # TODO: allow list of weights as well
    N_weights = 1 if weights is not None else 0

    # some sanity checks
    # TODO: replace this with a more robust function
    assert len(bins) == N_args
    for bin in bins:
        assert isinstance(bin, np.ndarray), "all bins must be numpy arrays"

    for a in args:
        # TODO: make this a more robust check
        assert a.name is not None, "all arrays must have a name"

    # we drop coords to simplify alignment
    if not keep_coords:
        args = [da.reset_coords(drop=True) for da in args]
    if N_weights:
        args += [weights.reset_coords(drop=True)]
    # explicitly broadcast so we understand what is going into apply_ufunc
    # (apply_ufunc might be doing this by itself again)
    args = list(xr.align(*args, join="exact"))

    # what happens if we skip this?
    # args = list(xr.broadcast(*args))
    a0 = args[0]
    a_coords = a0.coords

    # roll our own broadcasting
    # now manually expand the arrays
    all_dims = [d for a in args for d in a.dims]
    all_dims_ordered = list(OrderedDict.fromkeys(all_dims))
    args_expanded = []
    for a in args:
        expand_keys = [d for d in all_dims_ordered if d not in a.dims]
        a_expanded = a.expand_dims({k: 1 for k in expand_keys})
        args_expanded.append(a_expanded)

    # only transpose if necessary, to avoid creating unnecessary dask tasks
    args_transposed = []
    for a in args_expanded:
        if a.dims != all_dims_ordered:
            args_transposed.append(a.transpose(*all_dims_ordered))
        else:
            args.transposed.append(a)
    args_data = [a.data for a in args_transposed]

    if N_weights:
        weights_data = args_data.pop()
    else:
        weights_data = None

    if dim is not None:
        dims_to_keep = [d for d in all_dims_ordered if d not in dim]
        axis = [args_transposed[0].get_axis_num(d) for d in dim]
    else:
        dims_to_keep = []
        axis = None

    # Allow future warning for https://github.com/xgcm/xhistogram/pull/45
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message="Future versions of xhistogram\\.core\\.histogram will return",
            category=FutureWarning,
        )
        h_data = _histogram(
            *args_data,
            weights=weights_data,
            bins=bins,
            axis=axis,
            density=density,
            block_size=block_size
        )

    # create output dims
    new_dims = [a.name + bin_dim_suffix for a in args[:N_args]]
    output_dims = dims_to_keep + new_dims

    # create new coords
    bin_centers = [0.5 * (bin[:-1] + bin[1:]) for bin in bins]
    new_coords = {
        name: ((name,), bin_center, a.attrs)
        for name, bin_center, a in zip(new_dims, bin_centers, args)
    }

    # old coords associated with dims
    old_dim_coords = {name: a0[name] for name in dims_to_keep if name in a_coords}

    all_coords = {}
    all_coords.update(old_dim_coords)
    all_coords.update(new_coords)
    # add compatible coords
    if keep_coords:
        for c in a_coords:
            if c not in all_coords and set(a0[c].dims).issubset(output_dims):
                all_coords[c] = a0[c]

    output_name = "_".join(["histogram"] + [a.name for a in args[:N_args]])

    da_out = xr.DataArray(h_data, dims=output_dims, coords=all_coords, name=output_name)

    if density:
        # correct for overcounting the bins which weren't histogrammed along
        n_bins_bystander_dims = da_out.isel(**{bd: 0 for bd in new_dims}).size
        da_out = da_out * n_bins_bystander_dims

    return da_out

    # we need weights to be passed through apply_func's alignment algorithm,
    # so we include it as an arg, so we create a wrapper function to do so
    # this feels like a hack
    # def _histogram_wrapped(*args, **kwargs):
    #     alist = list(args)
    #     weights = [alist.pop() for n in range(N_weights)]
    #     if N_weights == 0:
    #         weights = None
    #     elif N_weights == 1:
    #         weights = weights[0] # squeeze
    #     return _histogram(*alist, weights=weights, **kwargs)



"""
Numpy API for xhistogram.
"""


import numpy as np
from functools import reduce
from .duck_array_ops import (
    digitize,
    bincount,
    reshape,
    ravel_multi_index,
    concatenate,
    broadcast_arrays,
)
import warnings


def _ensure_bins_is_a_list_of_arrays(bins, N_expected):
    if len(bins) == N_expected:
        return bins
    elif N_expected == 1:
        return [bins]
    else:
        raise ValueError("Can't figure out what to do with bins.")


def _bincount_2d(bin_indices, weights, N, hist_shapes):
    # a trick to apply bincount on an axis-by-axis basis
    # https://stackoverflow.com/questions/40591754/vectorizing-numpy-bincount
    # https://stackoverflow.com/questions/40588403/vectorized-searchsorted-numpy
    M = bin_indices.shape[0]
    if weights is not None:
        weights = weights.ravel()
    bin_indices_offset = (bin_indices + (N * np.arange(M)[:, None])).ravel()
    bc_offset = bincount(bin_indices_offset, weights=weights, minlength=N * M)
    final_shape = (M,) + tuple(hist_shapes)
    return bc_offset.reshape(final_shape)


def _bincount_loop(bin_indices, weights, N, hist_shapes, block_chunks):
    M = bin_indices.shape[0]
    assert sum(block_chunks) == M
    block_counts = []
    # iterate over chunks
    bounds = np.cumsum((0,) + block_chunks)
    for m_start, m_end in zip(bounds[:-1], bounds[1:]):
        bin_indices_block = bin_indices[m_start:m_end]
        weights_block = weights[m_start:m_end] if weights is not None else None
        bc_block = _bincount_2d(bin_indices_block, weights_block, N, hist_shapes)
        block_counts.append(bc_block)
    all_counts = concatenate(block_counts)
    final_shape = (bin_indices.shape[0],) + tuple(hist_shapes)
    return all_counts.reshape(final_shape)


def _determine_block_chunks(bin_indices, block_size):
    M, N = bin_indices.shape
    if block_size is None:
        return (M,)
    if block_size == "auto":
        try:
            # dask arrays - use the pre-existing chunks
            chunks = bin_indices.chunks
            return chunks[0]
        except AttributeError:
            # automatically pick a chunk size
            # this a a heueristic without much basis
            _MAX_CHUNK_SIZE = 10_000_000
            block_size = min(_MAX_CHUNK_SIZE // N, M)
    assert isinstance(block_size, int)
    num_chunks = M // block_size
    block_chunks = num_chunks * (block_size,)
    residual = M % block_size
    if residual:
        block_chunks += (residual,)
    assert sum(block_chunks) == M
    return block_chunks


def _dispatch_bincount(bin_indices, weights, N, hist_shapes, block_size=None):
    # block_chunks is like a dask chunk, a tuple that divides up the first
    # axis of bin_indices
    block_chunks = _determine_block_chunks(bin_indices, block_size)
    if len(block_chunks) == 1:
        # single global chunk, don't need a loop over chunks
        return _bincount_2d(bin_indices, weights, N, hist_shapes)
    else:
        return _bincount_loop(bin_indices, weights, N, hist_shapes, block_chunks)


def _histogram_2d_vectorized(
    *args, bins=None, weights=None, density=False, right=False, block_size=None
):
    """Calculate the histogram independently on each row of a 2D array"""

    N_inputs = len(args)
    a0 = args[0]

    # consistency checks for inputa
    for a, b in zip(args, bins):
        assert a.ndim == 2
        assert b.ndim == 1
        assert a.shape == a0.shape
    if weights is not None:
        assert weights.shape == a0.shape

    nrows, ncols = a0.shape
    nbins = [len(b) for b in bins]
    hist_shapes = [nb + 1 for nb in nbins]

    # a marginally faster implementation would be to use searchsorted,
    # like numpy histogram itself does
    # https://github.com/numpy/numpy/blob/9c98662ee2f7daca3f9fae9d5144a9a8d3cabe8c/numpy/lib/histograms.py#L864-L882
    # for now we stick with `digitize` because it's easy to understand how it works

    # Add small increment to the last bin edge to make the final bin right-edge inclusive
    # Note, this is the approach taken by sklearn, e.g.
    # https://github.com/scikit-learn/scikit-learn/blob/master/sklearn/calibration.py#L592
    # but a better approach would be to use something like _search_sorted_inclusive() in
    # numpy histogram. This is an additional motivation for moving to searchsorted
    bins = [np.concatenate((b[:-1], b[-1:] + 1e-8)) for b in bins]

    # the maximum possible value of of digitize is nbins
    # for right=False:
    #   - 0 corresponds to a < b[0]
    #   - i corresponds to bins[i-1] <= a < b[i]
    #   - nbins corresponds to a a >= b[1]
    each_bin_indices = [digitize(a, b) for a, b in zip(args, bins)]
    # product of the bins gives the joint distribution
    if N_inputs > 1:
        bin_indices = ravel_multi_index(each_bin_indices, hist_shapes)
    else:
        bin_indices = each_bin_indices[0]
    # total number of unique bin indices
    N = reduce(lambda x, y: x * y, hist_shapes)

    bin_counts = _dispatch_bincount(
        bin_indices, weights, N, hist_shapes, block_size=block_size
    )

    # just throw out everything outside of the bins, as np.histogram does
    # TODO: make this optional?
    slices = (slice(None),) + (N_inputs * (slice(1, -1),))
    bin_counts = bin_counts[slices]

    return bin_counts


def histogram(
    *args, bins=None, axis=None, weights=None, density=False, block_size="auto"
):
    """Histogram applied along specified axis / axes.

    Parameters
    ----------
    args : array_like
        Input data. The number of input arguments determines the dimensonality
        of the histogram. For example, two arguments prodocue a 2D histogram.
        All args must have the same size.
    bins :  int or array_like or a list of ints or arrays, optional
        If a list, there should be one entry for each item in ``args``.
        The bin specification:

          * If int, the number of bins for all arguments in ``args``.
          * If array_like, the bin edges for all arguments in ``args``.
          * If a list of ints, the number of bins  for every argument in ``args``.
          * If a list arrays, the bin edges for each argument in ``args``
            (required format for Dask inputs).
          * A combination [int, array] or [array, int], where int
            is the number of bins and array is the bin edges.

        When bin edges are specified, all but the last (righthand-most) bin include
        the left edge and exclude the right edge. The last bin includes both edges.

        A ``TypeError`` will be raised if ``args`` contains dask arrays and
        ``bins`` are not specified explicitly as a list of arrays.
    axis : None or int or tuple of ints, optional
        Axis or axes along which the histogram is computed. The default is to
        compute the histogram of the flattened array
    weights : array_like, optional
        An array of weights, of the same shape as `a`.  Each value in
        `a` only contributes its associated weight towards the bin count
        (instead of 1). If `density` is True, the weights are
        normalized, so that the integral of the density over the range
        remains 1.
    density : bool, optional
        If ``False``, the result will contain the number of samples in
        each bin. If ``True``, the result is the value of the
        probability *density* function at the bin, normalized such that
        the *integral* over the range is 1. Note that the sum of the
        histogram values will not be equal to 1 unless bins of unity
        width are chosen; it is not a probability *mass* function.
    block_size : int or 'auto', optional
        A parameter which governs the algorithm used to compute the histogram.
        Using a nonzero value splits the histogram calculation over the
        non-histogram axes into blocks of size ``block_size``, iterating over
        them with a loop (numpy inputs) or in parallel (dask inputs). If
        ``'auto'``, blocks will be determined either by the underlying dask
        chunks (dask inputs) or an experimental built-in heuristic (numpy inputs).

    Returns
    -------
    hist : array
        The values of the histogram.

    See Also
    --------
    numpy.histogram, numpy.bincount, numpy.digitize
    """

    # Future warning for https://github.com/xgcm/xhistogram/pull/45
    warnings.warn(
        "Future versions of xhistogram.core.histogram will return a "
        + "tuple containing arrays of the the histogram bins and the "
        + "histogram values, rather than just an array of the histogram "
        + "values. This API change will only affect users of "
        + "xhistogram.core. Users of xhistogram.xarray can ignore this "
        + "message.",
        FutureWarning,
    )

    a0 = args[0]
    ndim = a0.ndim

    if axis is not None:
        axis = np.atleast_1d(axis)
        assert axis.ndim == 1
        axis_normed = []
        for ax in axis:
            if ax >= 0:
                ax_positive = ax
            else:
                ax_positive = ndim + ax
            assert ax_positive < ndim, "axis must be less than ndim"
            axis_normed.append(ax_positive)
        axis = np.atleast_1d(axis_normed)

    do_full_array = (axis is None) or (set(axis) == set(range(a0.ndim)))
    if do_full_array:
        kept_axes_shape = None
    else:
        kept_axes_shape = tuple([a0.shape[i] for i in range(a0.ndim) if i not in axis])

    all_args = list(args)
    if weights is not None:
        all_args += [weights]
    all_args_broadcast = broadcast_arrays(*all_args)

    def reshape_input(a):
        if do_full_array:
            d = a.ravel()[None, :]
        else:
            # reshape the array to 2D
            # axis 0: preserved axis after histogram
            # axis 1: calculate histogram along this axis
            new_pos = tuple(range(-len(axis), 0))
            c = np.moveaxis(a, axis, new_pos)
            split_idx = c.ndim - len(axis)
            dims_0 = c.shape[:split_idx]
            assert dims_0 == kept_axes_shape
            dims_1 = c.shape[split_idx:]
            new_dim_0 = np.prod(dims_0)
            new_dim_1 = np.prod(dims_1)
            d = reshape(c, (new_dim_0, new_dim_1))
        return d

    all_args_reshaped = [reshape_input(a) for a in all_args_broadcast]

    if weights is not None:
        weights_reshaped = all_args_reshaped.pop()
    else:
        weights_reshaped = None

    n_inputs = len(all_args_reshaped)
    bins = _ensure_bins_is_a_list_of_arrays(bins, n_inputs)

    bin_counts = _histogram_2d_vectorized(
        *all_args_reshaped,
        bins=bins,
        weights=weights_reshaped,
        density=density,
        block_size=block_size,
    )

    if density:
        # Normalise by dividing by bin counts and areas such that all the
        # histogram data integrated over all dimensions = 1
        bin_widths = [np.diff(b) for b in bins]
        if n_inputs == 1:
            bin_areas = bin_widths[0]
        elif n_inputs == 2:
            bin_areas = np.outer(*bin_widths)
        else:
            # Slower, but N-dimensional logic
            bin_areas = np.prod(np.ix_(*bin_widths))

        h = bin_counts / bin_areas / bin_counts.sum()
    else:
        h = bin_counts

    if h.shape[0] == 1:
        assert do_full_array
        h = h.squeeze()
    else:
        final_shape = kept_axes_shape + h.shape[1:]
        h = reshape(h, final_shape)

    return h



import numpy as np

from itertools import combinations
import dask.array as dsa

from ..core import histogram
from .fixtures import empty_dask_array

import pytest


@pytest.mark.parametrize("density", [False, True])
@pytest.mark.parametrize("block_size", [None, 1, 2])
@pytest.mark.parametrize("axis", [1, None])
def test_histogram_results_1d(block_size, density, axis):
    nrows, ncols = 5, 20
    # Setting the random seed here prevents np.testing.assert_allclose
    # from failing beow. We should investigate this further.
    np.random.seed(2)
    data = np.random.randn(nrows, ncols)
    bins = np.linspace(-4, 4, 10)

    h = histogram(data, bins=bins, axis=axis, block_size=block_size, density=density)

    expected_shape = (nrows, len(bins) - 1) if axis == 1 else (len(bins) - 1,)
    assert h.shape == expected_shape

    # make sure we get the same thing as numpy.histogram
    if axis:
        expected = np.stack(
            [np.histogram(data[i], bins=bins, density=density)[0] for i in range(nrows)]
        )
    else:
        expected = np.histogram(data, bins=bins, density=density)[0]
    norm = nrows if (density and axis) else 1
    np.testing.assert_allclose(h, expected / norm)

    if density:
        widths = np.diff(bins)
        integral = np.sum(h * widths)
        np.testing.assert_allclose(integral, 1.0)


@pytest.mark.parametrize("block_size", [None, 1, 2])
def test_histogram_results_1d_weighted(block_size):
    nrows, ncols = 5, 20
    data = np.random.randn(nrows, ncols)
    bins = np.linspace(-4, 4, 10)
    h = histogram(data, bins=bins, axis=1, block_size=block_size)
    weights = 2 * np.ones_like(data)
    h_w = histogram(data, bins=bins, axis=1, weights=weights, block_size=block_size)
    np.testing.assert_array_equal(2 * h, h_w)


# @pytest.mark.skip(reason="Weight broadcasting on numpy arrays is not yet implemented")
@pytest.mark.parametrize("block_size", [None, 1, 2, "auto"])
def test_histogram_results_1d_weighted_broadcasting(block_size):
    nrows, ncols = 5, 20
    data = np.random.randn(nrows, ncols)
    bins = np.linspace(-4, 4, 10)
    h = histogram(data, bins=bins, axis=1, block_size=block_size)
    weights = 2 * np.ones((1, ncols))
    h_w = histogram(data, bins=bins, axis=1, weights=weights, block_size=block_size)
    np.testing.assert_array_equal(2 * h, h_w)


@pytest.mark.parametrize("block_size", [None, 1, 2])
def test_histogram_right_edge(block_size):
    """Test that last bin is both left- and right-edge inclusive as it
    is for numpy.histogram
    """
    nrows, ncols = 5, 20
    data = np.ones((nrows, ncols))
    bins = np.array([0, 0.5, 1])  # All data at rightmost edge

    h = histogram(data, bins=bins, axis=1, block_size=block_size)
    assert h.shape == (nrows, len(bins) - 1)

    # make sure we get the same thing as histogram (all data in the last bin)
    hist, _ = np.histogram(data, bins=bins)
    np.testing.assert_array_equal(hist, h.sum(axis=0))

    # now try with no axis
    h_na = histogram(data, bins=bins, block_size=block_size)
    np.testing.assert_array_equal(hist, h_na)


def test_histogram_results_2d():
    nrows, ncols = 5, 20
    data_a = np.random.randn(nrows, ncols)
    data_b = np.random.randn(nrows, ncols)
    nbins_a = 9
    bins_a = np.linspace(-4, 4, nbins_a + 1)
    nbins_b = 10
    bins_b = np.linspace(-4, 4, nbins_b + 1)

    h = histogram(data_a, data_b, bins=[bins_a, bins_b])
    assert h.shape == (nbins_a, nbins_b)

    hist, _, _ = np.histogram2d(data_a.ravel(), data_b.ravel(), bins=[bins_a, bins_b])
    np.testing.assert_array_equal(hist, h)


def test_histogram_results_2d_density():
    nrows, ncols = 5, 20
    data_a = np.random.randn(nrows, ncols)
    data_b = np.random.randn(nrows, ncols)
    nbins_a = 9
    bins_a = np.linspace(-4, 4, nbins_a + 1)
    nbins_b = 10
    bins_b = np.linspace(-4, 4, nbins_b + 1)

    h = histogram(data_a, data_b, bins=[bins_a, bins_b], density=True)
    assert h.shape == (nbins_a, nbins_b)

    hist, _, _ = np.histogram2d(
        data_a.ravel(), data_b.ravel(), bins=[bins_a, bins_b], density=True
    )
    np.testing.assert_allclose(hist, h)

    # check integral is 1
    widths_a = np.diff(bins_a)
    widths_b = np.diff(bins_b)
    areas = np.outer(widths_a, widths_b)
    integral = np.sum(hist * areas)
    np.testing.assert_allclose(integral, 1.0)


def test_histogram_results_3d_density():
    nrows, ncols = 5, 20
    data_a = np.random.randn(nrows, ncols)
    data_b = np.random.randn(nrows, ncols)
    data_c = np.random.randn(nrows, ncols)
    nbins_a = 9
    bins_a = np.linspace(-4, 4, nbins_a + 1)
    nbins_b = 10
    bins_b = np.linspace(-4, 4, nbins_b + 1)
    nbins_c = 9
    bins_c = np.linspace(-4, 4, nbins_c + 1)

    h = histogram(data_a, data_b, data_c, bins=[bins_a, bins_b, bins_c], density=True)

    assert h.shape == (nbins_a, nbins_b, nbins_c)

    hist, _ = np.histogramdd(
        (data_a.ravel(), data_b.ravel(), data_c.ravel()),
        bins=[bins_a, bins_b, bins_c],
        density=True,
    )

    np.testing.assert_allclose(hist, h)

    # check integral is 1
    widths_a = np.diff(bins_a)
    widths_b = np.diff(bins_b)
    widths_c = np.diff(bins_c)
    areas = np.einsum("i,j,k", widths_a, widths_b, widths_c)
    integral = np.sum(hist * areas)
    np.testing.assert_allclose(integral, 1.0)


@pytest.mark.parametrize("block_size", [None, 5, "auto"])
@pytest.mark.parametrize("use_dask", [False, True])
def test_histogram_shape(use_dask, block_size):
    """These tests just verify that arrays with the right shape come out.
    They don't verify correctness."""

    shape = 10, 15, 12, 20
    if use_dask:
        b = empty_dask_array(shape, chunks=(1,) + shape[1:])
    else:
        b = np.random.randn(*shape)
    bins = np.linspace(-4, 4, 27)

    # no axis
    c = histogram(b, bins=bins, block_size=block_size)
    assert c.shape == (len(bins) - 1,)
    # same thing
    for axis in [(0, 1, 2, 3), (0, 1, 3, 2), (3, 2, 1, 0), (3, 2, 0, 1)]:
        c = histogram(b, bins=bins, axis=axis)
        assert c.shape == (len(bins) - 1,)
        if use_dask:
            assert isinstance(c, dsa.Array)

    # scalar axis (check positive and negative)
    for axis in list(range(4)) + list(range(-1, -5, -1)):
        c = histogram(b, bins=bins, axis=axis, block_size=block_size)
        shape = list(b.shape)
        del shape[axis]
        expected_shape = tuple(shape) + (len(bins) - 1,)
        assert c.shape == expected_shape
        if use_dask:
            assert isinstance(c, dsa.Array)

    # two axes
    for i, j in combinations(range(4), 2):
        axis = (i, j)
        c = histogram(b, bins=bins, axis=axis, block_size=block_size)
        shape = list(b.shape)
        partial_shape = [shape[k] for k in range(b.ndim) if k not in axis]
        expected_shape = tuple(partial_shape) + (len(bins) - 1,)
        assert c.shape == expected_shape
        if use_dask:
            assert isinstance(c, dsa.Array)



"""Compatibility module defining operations on duck numpy-arrays.
Shamelessly copied from xarray."""

import numpy as np

try:
    import dask.array as dsa

    has_dask = True
except ImportError:
    has_dask = False


def _dask_or_eager_func(name, eager_module=np, list_of_args=False, n_array_args=1):
    """Create a function that dispatches to dask for dask array inputs."""
    if has_dask:

        def f(*args, **kwargs):
            dispatch_args = args[0] if list_of_args else args
            if any(isinstance(a, dsa.Array) for a in dispatch_args[:n_array_args]):
                module = dsa
            else:
                module = eager_module
            return getattr(module, name)(*args, **kwargs)

    else:

        def f(*args, **kwargs):
            return getattr(eager_module, name)(*args, **kwargs)

    return f


digitize = _dask_or_eager_func("digitize")
bincount = _dask_or_eager_func("bincount")
reshape = _dask_or_eager_func("reshape")
concatenate = _dask_or_eager_func("concatenate", list_of_args=True)
broadcast_arrays = _dask_or_eager_func("broadcast_arrays")
ravel_multi_index = _dask_or_eager_func("ravel_multi_index")



#!/usr/bin/env python
import versioneer
from setuptools import setup, find_packages

DISTNAME = "xhistogram"
LICENSE = "MIT"
AUTHOR = "xhistogram Developers"
AUTHOR_EMAIL = "rpa@ldeo.columbia.edu"
URL = "https://github.com/xgcm/xhistogram"
CLASSIFIERS = [
    "Development Status :: 4 - Beta",
    "License :: OSI Approved :: Apache Software License",
    "Operating System :: OS Independent",
    "Intended Audience :: Science/Research",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Topic :: Scientific/Engineering",
]

INSTALL_REQUIRES = ["xarray>=0.12.0", "dask", "numpy>=1.16"]
PYTHON_REQUIRES = ">=3.7"

DESCRIPTION = "Fast, flexible, label-aware histograms for numpy and xarray"


def readme():
    with open("README.rst") as f:
        return f.read()


setup(
    name=DISTNAME,
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    license=LICENSE,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    classifiers=CLASSIFIERS,
    description=DESCRIPTION,
    long_description=readme(),
    install_requires=INSTALL_REQUIRES,
    python_requires=PYTHON_REQUIRES,
    url=URL,
    packages=find_packages(),
)
